@@ -1,6 +1,6 @@
 /*
 **  GNU Pth - The GNU Portable Threads
-**  Copyright (c) 1999 Ralf S. Engelschall <rse@engelschall.com>
+**  Copyright (c) 1999-2000 Ralf S. Engelschall <rse@engelschall.com>
 **
 **  This file is part of GNU Pth, a non-preemptive thread scheduling
 **  library which can be found at http://www.gnu.org/software/pth/.
@@ -46,19 +46,23 @@ struct pth_event_st {
         struct { pth_mutex_t *mutex; }                            MUTEX;
         struct { pth_cond_t *cond; }                              COND;
         struct { pth_t tid; }                                     TID;
-        struct { pid_t pid; int *status; int flags; }             PID;
-        struct { int (*func)(void*); void *func_arg; }            FUNC;
+        struct { int (*func)(void *); void *arg; pth_time_t tv; } FUNC;
     } ev_args;
 };
 
 #endif /* cpp */
 
-/* event structure constructor */
+/* event structure destructor */
 static void pth_event_destructor(void *vp)
 {
-    pth_event_free((pth_event_t)vp, PTH_FREE_ALL);
+    /* free this single(!) event. That it is just a single event is a
+       requirement for pth_event(PTH_MODE_STATIC, ...), or else we would
+       get into horrible trouble on asychronous cleanups */
+    pth_event_free((pth_event_t)vp, PTH_FREE_THIS);
     return;
 }
+
+/* event structure constructor */
 pth_event_t pth_event(unsigned long spec, ...)
 {
     pth_event_t ev;
@@ -66,7 +70,7 @@ pth_event_t pth_event(unsigned long spec, ...)
     va_list ap;
 
     va_start(ap, spec);
-    
+
     /* allocate new or reuse static or supplied event structure */
     if (spec & PTH_MODE_REUSE) {
         /* reuse supplied event structure */
@@ -186,20 +190,13 @@ pth_event_t pth_event(unsigned long spec, ...)
         ev->ev_goal = goal;
         ev->ev_args.TID.tid = tid;
     }
-    else if (spec & PTH_EVENT_PID) {
-        /* process id event */
-        ev->ev_type = PTH_EVENT_PID;
-        ev->ev_goal = (int)(spec & (PTH_UNTIL_OCCURRED));
-        ev->ev_args.PID.pid    = va_arg(ap, pid_t);
-        ev->ev_args.PID.status = va_arg(ap, int *);
-        ev->ev_args.PID.flags  = va_arg(ap, int);
-    }
     else if (spec & PTH_EVENT_FUNC) {
         /* custom function event */
         ev->ev_type = PTH_EVENT_FUNC;
         ev->ev_goal = (int)(spec & (PTH_UNTIL_OCCURRED));
-        ev->ev_args.FUNC.func     = (int (*)(void *))va_arg(ap, void *);
-        ev->ev_args.FUNC.func_arg = va_arg(ap, void *);
+        ev->ev_args.FUNC.func  = (int (*)(void *))va_arg(ap, void *);
+        ev->ev_args.FUNC.arg   = va_arg(ap, void *);
+        ev->ev_args.FUNC.tv    = va_arg(ap, pth_time_t);
     }
     else
         return_errno(NULL, EINVAL);
@@ -265,21 +262,14 @@ int pth_event_extract(pth_event_t ev, ...)
         pth_t *tid = va_arg(ap, pth_t *);
         *tid = ev->ev_args.TID.tid;
     }
-    else if (ev->ev_type & PTH_EVENT_PID) {
-        /* process id event */
-        pid_t *pid   = va_arg(ap, pid_t *);
-        int **status = va_arg(ap, int **);
-        int *flags   = va_arg(ap, int *);
-        *pid    = ev->ev_args.PID.pid;
-        *status = ev->ev_args.PID.status;
-        *flags  = ev->ev_args.PID.flags;
-    }
     else if (ev->ev_type & PTH_EVENT_FUNC) {
         /* custom function event */
-        void **func = va_arg(ap, void **);
-        void **arg  = va_arg(ap, void **);
+        void **func    = va_arg(ap, void **);
+        void **arg     = va_arg(ap, void **);
+        pth_time_t *tv = va_arg(ap, pth_time_t *);
         *func = (int (**)(void *))ev->ev_args.FUNC.func;
-        *arg  = ev->ev_args.FUNC.func_arg;
+        *arg  = ev->ev_args.FUNC.arg;
+        *tv   = ev->ev_args.FUNC.tv;
     }
     else
         return_errno(FALSE, EINVAL);
@@ -370,7 +360,7 @@ int pth_event_free(pth_event_t ev, int mode)
 {
     pth_event_t evc;
     pth_event_t evn;
-    
+
     if (ev == NULL)
         return_errno(FALSE, EINVAL);
     if (mode == PTH_FREE_THIS) {
@@ -404,14 +394,14 @@ int pth_wait(pth_event_t ev_ring)
     ev = ev_ring;
     do {
         ev->ev_occurred = FALSE;
-        pth_debug2("pth_wait: waiting event 0x%lx", (unsigned long)ev);
+        pth_debug2("pth_wait: waiting on event 0x%lx", (unsigned long)ev);
         ev = ev->ev_next;
     } while (ev != ev_ring);
 
     /* link event ring to current thread */
     pth_current->events = ev_ring;
 
-    /* move thread into waiting state 
+    /* move thread into waiting state
        and transfer control to scheduler */
     pth_current->state = PTH_STATE_WAITING;
     pth_yield(NULL);
@@ -433,6 +423,6 @@ int pth_wait(pth_event_t ev_ring)
 
     /* leave to current thread with number of occurred events */
     pth_debug2("pth_wait: leave to thread \"%s\"", pth_current->name);
-    return occurred; 
+    return occurred;
 }
 
