@@ -1,6 +1,6 @@
 /*
 **  GNU Pth - The GNU Portable Threads
-**  Copyright (c) 1999 Ralf S. Engelschall <rse@engelschall.com>
+**  Copyright (c) 1999-2000 Ralf S. Engelschall <rse@engelschall.com>
 **
 **  This file is part of GNU Pth, a non-preemptive thread scheduling
 **  library which can be found at http://www.gnu.org/software/pth/.
@@ -51,7 +51,7 @@ int pth_init(void)
     /* support for implicit initialization calls
        and to prevent multiple explict initialization, too */
     if (pth_initialized)
-        return FALSE;
+        return_errno(FALSE, EPERM);
     else
         pth_initialized = TRUE;
 
@@ -69,8 +69,13 @@ int pth_init(void)
     pth_attr_set(t_attr, PTH_ATTR_STACK_SIZE,   32*1024);
     pth_attr_set(t_attr, PTH_ATTR_STACK_ADDR,   NULL);
     pth_sched = pth_spawn(t_attr, pth_scheduler, NULL);
-    if (pth_sched == NULL)
+    if (pth_sched == NULL) {
+        errno_shield { 
+            pth_attr_destroy(t_attr);
+            pth_scheduler_kill();
+        }
         return FALSE;
+    }
 
     /* spawn a thread for the main program */
     pth_attr_set(t_attr, PTH_ATTR_PRIO,         PTH_PRIO_STD);
@@ -80,11 +85,16 @@ int pth_init(void)
     pth_attr_set(t_attr, PTH_ATTR_STACK_SIZE,   0 /* special */);
     pth_attr_set(t_attr, PTH_ATTR_STACK_ADDR,   NULL);
     pth_main = pth_spawn(t_attr, (void *(*)(void *))(-1), NULL);
-    pth_attr_destroy(t_attr);
-    if (pth_main == NULL)
+    if (pth_main == NULL) {
+        errno_shield { 
+            pth_attr_destroy(t_attr);
+            pth_scheduler_kill();
+        }
         return FALSE;
+    }
+    pth_attr_destroy(t_attr);
 
-    /* 
+    /*
      * The first time we've to manually switch into the scheduler to start
      * threading. Because at this time the only non-scheduler thread is the
      * "main thread" we will come back immediately. We've to also initialize
@@ -103,8 +113,9 @@ int pth_init(void)
 int pth_kill(void)
 {
     if (pth_current != pth_main) 
-        return_errno(FALSE, EINVAL);
+        return_errno(FALSE, EPERM);
     pth_debug1("pth_kill: enter");
+    pth_thread_cleanup(pth_main);
     pth_scheduler_kill();
     pth_initialized = FALSE;
     pth_tcb_free(pth_sched);
@@ -122,30 +133,36 @@ long pth_ctrl(unsigned long query, ...)
     rc = 0;
     va_start(ap, query);
     if (query & PTH_CTRL_GETTHREADS) {
-        if (query & PTH_CTRL_GETTHREADS_NEW) 
+        if (query & PTH_CTRL_GETTHREADS_NEW)
             rc += pth_pqueue_elements(&pth_NQ);
-        if (query & PTH_CTRL_GETTHREADS_READY) 
+        if (query & PTH_CTRL_GETTHREADS_READY)
             rc += pth_pqueue_elements(&pth_RQ);
-        if (query & PTH_CTRL_GETTHREADS_RUNNING) 
+        if (query & PTH_CTRL_GETTHREADS_RUNNING)
             rc += 1; /* pth_current only */
-        if (query & PTH_CTRL_GETTHREADS_WAITING) 
+        if (query & PTH_CTRL_GETTHREADS_WAITING)
             rc += pth_pqueue_elements(&pth_WQ);
-        if (query & PTH_CTRL_GETTHREADS_DEAD) 
+        if (query & PTH_CTRL_GETTHREADS_SUSPENDED) 
+            rc += pth_pqueue_elements(&pth_SQ);
+        if (query & PTH_CTRL_GETTHREADS_DEAD)
             rc += pth_pqueue_elements(&pth_DQ);
     }
     else if (query & PTH_CTRL_GETAVLOAD) {
-        float *pload = va_arg(ap, float *); 
+        float *pload = va_arg(ap, float *);
         *pload = pth_loadval;
     }
     else if (query & PTH_CTRL_GETPRIO) {
-        pth_t t = va_arg(ap, pth_t); 
+        pth_t t = va_arg(ap, pth_t);
         rc = t->prio;
     }
     else if (query & PTH_CTRL_GETNAME) {
-        pth_t t = va_arg(ap, pth_t); 
+        pth_t t = va_arg(ap, pth_t);
         rc = (long)t->name;
     }
-    else 
+    else if (query & PTH_CTRL_DUMPSTATE) {
+        FILE *fp = va_arg(ap, FILE *);
+        pth_dumpstate(fp);
+    }
+    else
         rc = -1;
     va_end(ap);
     if (rc == -1)
@@ -163,9 +180,9 @@ static void pth_spawn_trampoline(void)
     /* and do an implicit exit of the tread with the result value */
     pth_exit(data);
     /* no return! */
-    abort(); 
+    abort();
 }
-pth_t pth_spawn(pth_attr_t attr, void *(*func)(void*), void *arg)
+pth_t pth_spawn(pth_attr_t attr, void *(*func)(void *), void *arg)
 {
     pth_t t;
     int stacksize;
@@ -201,7 +218,9 @@ pth_t pth_spawn(pth_attr_t attr, void *(*func)(void*), void *arg)
         t->prio        = pth_current->prio;
         t->joinable    = pth_current->joinable;
         t->cancelstate = pth_current->cancelstate;
-        sprintf(t->name, "%s.child/%x", pth_current->name, (unsigned int)time(NULL));
+        sprintf(t->name, "%s.child@%d=0x%lx", 
+                pth_current->name, (unsigned int)time(NULL), 
+                (unsigned long)pth_current);
     }
     else {
         /* defaults */
@@ -230,11 +249,11 @@ pth_t pth_spawn(pth_attr_t attr, void *(*func)(void*), void *arg)
 
     /* initialize join argument */
     t->join_arg = NULL;
-    
+
     /* initialize thread specific storage */
     t->data_value = NULL;
     t->data_count = 0;
-    
+
     /* initialize cancellaton stuff */
     t->cancelreq   = FALSE;
     t->cleanups    = NULL;
@@ -243,11 +262,14 @@ pth_t pth_spawn(pth_attr_t attr, void *(*func)(void*), void *arg)
     pth_ring_init(&(t->mutexring), NULL);
 
     /* initialize the machine context of this new thread */
-    if (t->stacksize > 0) /* the "main thread" (indicated by == 0) is special! */
+    if (t->stacksize > 0) { /* the "main thread" (indicated by == 0) is special! */
         if (!pth_mctx_set(&t->mctx, pth_spawn_trampoline,
-                          t->stack, ((char *)t->stack+t->stacksize)))
+                          t->stack, ((char *)t->stack+t->stacksize))) {
+            errno_shield { pth_tcb_free(t); }
             return NULL;
-    
+        }
+    }
+
     /* finally insert it into the "new queue" where
        the scheduler will pick it up for dispatching */
     if (func != pth_scheduler) {
@@ -257,7 +279,7 @@ pth_t pth_spawn(pth_attr_t attr, void *(*func)(void*), void *arg)
 
     pth_debug1("pth_spawn: leave");
 
-    /* the returned thread id is just the pointer 
+    /* the returned thread id is just the pointer
        to the thread control block... */
     return t;
 }
@@ -275,21 +297,11 @@ int pth_raise(pth_t t, int sig)
 
     if (t == NULL || t == pth_current || (sig < 0 || sig > PTH_NSIG))
         return_errno(FALSE, EINVAL);
-    if (sig == 0) {
-        /* 
-         * test whether thread exists
-         */
-        if (!pth_pqueue_contains(&pth_NQ, t))
-            if (!pth_pqueue_contains(&pth_RQ, t))
-                if (!pth_pqueue_contains(&pth_WQ, t))
-                    if (!pth_pqueue_contains(&pth_DQ, t))
-                        return_errno(FALSE, ESRCH); /* not found */
-        return TRUE;
-    }
+    if (sig == 0)
+        /* just test whether thread exists */
+        return pth_thread_exists(t);
     else {
-        /* 
-         * raise signal for thread
-         */
+        /* raise signal for thread */
         if (sigaction(sig, NULL, &sa) != 0)
             return FALSE;
         if (sa.sa_handler == SIG_IGN)
@@ -303,6 +315,35 @@ int pth_raise(pth_t t, int sig)
     }
 }
 
+/* check whether a thread exists */
+intern int pth_thread_exists(pth_t t)
+{
+    if (!pth_pqueue_contains(&pth_NQ, t))
+        if (!pth_pqueue_contains(&pth_RQ, t))
+            if (!pth_pqueue_contains(&pth_WQ, t))
+                if (!pth_pqueue_contains(&pth_SQ, t))
+                    if (!pth_pqueue_contains(&pth_DQ, t))
+                        return_errno(FALSE, ESRCH); /* not found */
+    return TRUE;
+}
+
+/* cleanup a particular thread */
+intern void pth_thread_cleanup(pth_t thread)
+{
+    /* run the cleanup handlers */
+    if (thread->cleanups != NULL)
+        pth_cleanup_popall(thread, TRUE);
+
+    /* run the specific data destructors */
+    if (thread->data_value != NULL)
+        pth_key_destroydata(thread);
+
+    /* release still acquired mutex variables */
+    pth_mutex_releaseall(thread);
+
+    return;
+}
+
 /* terminates the current thread */
 static int pth_exit_cb(void *arg)
 {
@@ -314,6 +355,7 @@ static int pth_exit_cb(void *arg)
     rc += pth_pqueue_elements(&pth_NQ);
     rc += pth_pqueue_elements(&pth_RQ);
     rc += pth_pqueue_elements(&pth_WQ);
+    rc += pth_pqueue_elements(&pth_SQ);
     rc += pth_pqueue_elements(&pth_DQ);
     if (rc == 1 /* just our main thread */)
         return TRUE;
@@ -326,42 +368,41 @@ void pth_exit(void *value)
 
     pth_debug2("pth_exit: marking thread \"%s\" as dead", pth_current->name);
 
-    /* main thread is special */
+    /* main thread is special:
+       wait until it is the last thread */
     if (pth_current == pth_main) {
-        /* wait until we're the last thread */
         ev = pth_event(PTH_EVENT_FUNC, pth_exit_cb);
         pth_wait(ev);
-        /* exit the _process_ [double-cast to avoid warnings because of size] */
-        pth_kill();
-        exit((int)((long)value)); 
+        pth_event_free(ev, PTH_FREE_THIS);
     }
 
-    /* run the cleanup handlers */
-    if (pth_current->cleanups != NULL)
-        pth_cleanup_popall(pth_current, TRUE);
-
-    /* run the per-thread specific data destructors */
-    if (pth_current->data_value != NULL)
-        pth_key_destroydata(pth_current);
-
-    /* release still acquired mutex variables */
-    pth_mutex_releaseall(pth_current);
+    /* execute cleanups */
+    pth_thread_cleanup(pth_current);
 
     /* mark the current thread as dead, so the scheduler removes us */
     pth_current->join_arg = value;
     pth_current->state = PTH_STATE_DEAD;
 
-    /* 
-     * Now we explicitly switch into the scheduler and let it
-     * reap the current thread structure; we can't free it here,
-     * or we'd be running on a stack which malloc() regards as
-     * free memory, which would be a somewhat perilous situation.
-     */
-    pth_debug2("pth_exit: switching from thread \"%s\" to scheduler", pth_current->name);
-    pth_mctx_switch(&pth_current->mctx, &pth_sched->mctx);
-
-    /* not reached! */
-    abort();
+    if (pth_current != pth_main) {
+        /*
+         * Now we explicitly switch into the scheduler and let it
+         * reap the current thread structure; we can't free it here,
+         * or we'd be running on a stack which malloc() regards as
+         * free memory, which would be a somewhat perilous situation.
+         */
+        pth_debug2("pth_exit: switching from thread \"%s\" to scheduler", pth_current->name);
+        pth_mctx_switch(&pth_current->mctx, &pth_sched->mctx);
+        abort(); /* not reached! */
+    }
+    else {
+        /* 
+         * main thread is special: exit the _process_ 
+         * [double-cast to avoid warnings because of size] 
+         */ 
+        pth_kill();
+        exit((int)((long)value));
+        abort(); /* not reached! */
+    }
 }
 
 /* waits for the termination of the specified thread */
@@ -370,6 +411,7 @@ int pth_join(pth_t tid, void **value)
     pth_event_t ev;
     static pth_key_t ev_key = PTH_KEY_INIT;
 
+    pth_debug2("pth_join: joining thread \"%s\"", tid == NULL ? "-ANY-" : tid->name);
     if (tid == pth_current)
         return_errno(FALSE, EDEADLK);
     if (tid != NULL && !tid->joinable)
@@ -407,7 +449,7 @@ int pth_yield(pth_t to)
         if (to == NULL)
             return_errno(FALSE, EINVAL);
     }
-    
+
     /* give a favored thread maximum priority in ready queue */
     if (to != NULL)
         pth_pqueue_favorite(&pth_RQ, to);
@@ -425,6 +467,52 @@ int pth_yield(pth_t to)
     return TRUE;
 }
 
+/* suspend a thread until its again manually resumed */
+int pth_suspend(pth_t t)
+{
+    pth_pqueue_t *q;
+
+    if (t == NULL)
+        return_errno(FALSE, EINVAL);
+    if (t == pth_sched || t == pth_current)
+        return_errno(FALSE, EPERM);
+    switch (t->state) {
+        case PTH_STATE_NEW:     q = &pth_NQ; break;
+        case PTH_STATE_READY:   q = &pth_RQ; break;
+        case PTH_STATE_WAITING: q = &pth_WQ; break;
+        default:                q = NULL;
+    }
+    if (q == NULL)
+        return_errno(FALSE, EPERM);
+    if (!pth_pqueue_contains(q, t))
+        return_errno(FALSE, ESRCH);
+    pth_pqueue_delete(q, t);
+    pth_pqueue_insert(&pth_SQ, PTH_PRIO_STD, t);
+    return TRUE;
+}
+
+/* resume a previously suspended thread */
+int pth_resume(pth_t t)
+{
+    pth_pqueue_t *q;
+
+    if (t == NULL)
+        return_errno(FALSE, EINVAL);
+    if (t == pth_sched || t == pth_current)
+        return_errno(FALSE, EPERM);
+    if (!pth_pqueue_contains(&pth_SQ, t))
+        return_errno(FALSE, EPERM);
+    pth_pqueue_delete(&pth_SQ, t);
+    switch (t->state) {
+        case PTH_STATE_NEW:     q = &pth_NQ; break;
+        case PTH_STATE_READY:   q = &pth_RQ; break;
+        case PTH_STATE_WAITING: q = &pth_WQ; break;
+        default:                q = NULL;
+    }
+    pth_pqueue_insert(q, PTH_PRIO_STD, t);
+    return TRUE;
+}
+
 /* switch a filedescriptor's I/O mode */
 int pth_fdmode(int fd, int newmode)
 {
@@ -436,7 +524,7 @@ int pth_fdmode(int fd, int newmode)
         oldmode = PTH_FDMODE_ERROR;
     else if (fdmode & O_NONBLOCKING)
         oldmode = PTH_FDMODE_NONBLOCK;
-    else 
+    else
         oldmode = PTH_FDMODE_BLOCK;
 
     /* set new mode (usually expensive) */
