@@ -1,6 +1,6 @@
 /*
 **  GNU Pth - The GNU Portable Threads
-**  Copyright (c) 1999 Ralf S. Engelschall <rse@engelschall.com>
+**  Copyright (c) 1999-2000 Ralf S. Engelschall <rse@engelschall.com>
 **
 **  This file is part of GNU Pth, a non-preemptive thread scheduling
 **  library which can be found at http://www.gnu.org/software/pth/.
@@ -33,6 +33,7 @@ intern pth_t        pth_current;    /* the currently running thread          */
 intern pth_pqueue_t pth_NQ;         /* queue of new threads                  */
 intern pth_pqueue_t pth_RQ;         /* queue of threads ready to run         */
 intern pth_pqueue_t pth_WQ;         /* queue of threads waiting for an event */
+intern pth_pqueue_t pth_SQ;         /* queue of suspended threads            */
 intern pth_pqueue_t pth_DQ;         /* queue of terminated threads           */
 intern float        pth_loadval;    /* average scheduler load value          */
 
@@ -50,7 +51,7 @@ intern void pth_scheduler_init(void)
 {
     /* create the internal signal pipe */
     if (pipe(pth_sigpipe) == -1) {
-        fprintf(stderr, "**Pth** INIT: Cannot create internal pipe: %s\n", 
+        fprintf(stderr, "**Pth** INIT: Cannot create internal pipe: %s\n",
                 strerror(errno));
         abort();
     }
@@ -65,6 +66,7 @@ intern void pth_scheduler_init(void)
     pth_pqueue_init(&pth_NQ);
     pth_pqueue_init(&pth_RQ);
     pth_pqueue_init(&pth_WQ);
+    pth_pqueue_init(&pth_SQ);
     pth_pqueue_init(&pth_DQ);
 
     /* initialize load support */
@@ -93,6 +95,11 @@ intern void pth_scheduler_drop(void)
         pth_tcb_free(t);
     pth_pqueue_init(&pth_WQ);
 
+    /* clear the suspend queue */
+    while ((t = pth_pqueue_delmax(&pth_SQ)) != NULL);
+        pth_tcb_free(t);
+    pth_pqueue_init(&pth_SQ);
+
     /* clear the dead queue */
     while ((t = pth_pqueue_delmax(&pth_DQ)) != NULL);
         pth_tcb_free(t);
@@ -110,9 +117,9 @@ intern void pth_scheduler_kill(void)
     close(pth_sigpipe[0]);
     close(pth_sigpipe[1]);
     return;
-} 
+}
 
-/* 
+/*
  * Update the average scheduler load.
  *
  * This is called on every context switch, but we have to adjust the
@@ -152,7 +159,7 @@ intern void *pth_scheduler(void *dummy)
     int sig;
     pth_t t;
 
-    /* 
+    /*
      * bootstrapping
      */
     pth_debug1("pth_scheduler: bootstrapping");
@@ -167,7 +174,7 @@ intern void *pth_scheduler(void *dummy)
     /* initialize the snapshot time for bootstrapping the loop */
     pth_time_set(&snapshot, PTH_TIME_NOW);
 
-    /* 
+    /*
      * endless scheduler loop
      */
     while (1) {
@@ -220,7 +227,7 @@ intern void *pth_scheduler(void *dummy)
          * Set running start time for new thread
          * and perform a context switch to it
          */
-        pth_debug3("pth_scheduler: switching to thread 0x%lx (\"%s\")", 
+        pth_debug3("pth_scheduler: switching to thread 0x%lx (\"%s\")",
                    (unsigned long)pth_current, pth_current->name);
 
         /* update thread times */
@@ -285,9 +292,9 @@ intern void *pth_scheduler(void *dummy)
             if (*pth_current->stackguard != 0xDEAD) {
                 pth_debug3("pth_scheduler: stack overflow detected for thread 0x%lx (\"%s\")",
                            (unsigned long)pth_current, pth_current->name);
-                /* 
-                 * when the application doesn't catch SIGSEGVs we terminate
-                 * manually with a SIGSEGV now but output a reasonable message
+                /*
+                 * if the application doesn't catch SIGSEGVs, we terminate
+                 * manually with a SIGSEGV now, but output a reasonable message.
                  */
                 if (sigaction(SIGSEGV, NULL, &sa) == 0) {
                     if (sa.sa_handler == SIG_DFL) {
@@ -300,7 +307,7 @@ intern void *pth_scheduler(void *dummy)
                         abort();
                     }
                 }
-                /* 
+                /*
                  * else we terminate the thread only and send us a SIGSEGV
                  * which allows the application to handle the situation...
                  */
@@ -323,11 +330,11 @@ intern void *pth_scheduler(void *dummy)
         }
 
         /*
-         * When thread wants to wait for an event 
+         * When thread wants to wait for an event
          * move it to waiting queue now
          */
         if (pth_current != NULL && pth_current->state == PTH_STATE_WAITING) {
-            pth_debug2("pth_scheduler: moving thread \"%s\" to waiting queue", 
+            pth_debug2("pth_scheduler: moving thread \"%s\" to waiting queue",
                        pth_current->name);
             pth_pqueue_insert(&pth_WQ, pth_current->prio, pth_current);
             pth_current = NULL;
@@ -344,8 +351,8 @@ intern void *pth_scheduler(void *dummy)
 
         /*
          * Manage the events in the waiting queue, i.e. decide whether their
-         * events occurred and move them to the ready queue. But wait only when
-         * we've already no new or ready threads.
+         * events occurred and move them to the ready queue. But wait only if
+         * we have already no new or ready threads.
          */
         if (   pth_pqueue_elements(&pth_RQ) == 0
             && pth_pqueue_elements(&pth_NQ) == 0)
@@ -353,7 +360,8 @@ intern void *pth_scheduler(void *dummy)
         else
             pth_sched_eventmanager(&snapshot, TRUE  /* poll */);
     }
-    /* NOT REACHED */
+
+    /* NOTREACHED */
     return NULL;
 }
 
@@ -381,14 +389,19 @@ intern void pth_sched_eventmanager(pth_time_t *now, int dopoll)
     struct sigaction sa;
     struct sigaction osa[1+PTH_NSIG];
     char minibuf[128];
+    int loop_repeat;
     int fdmax;
     int rc;
     int sig;
     int n;
 
-    pth_debug2("pth_sched_eventmanager: enter in %s mode", 
+    pth_debug2("pth_sched_eventmanager: enter in %s mode",
                dopoll ? "polling" : "waiting");
 
+    /* entry point for internal looping in event handling */
+    loop_entry:
+    loop_repeat = FALSE;
+
     /* initialize fd sets */
     FD_ZERO(&rfds);
     FD_ZERO(&wfds);
@@ -408,9 +421,9 @@ intern void pth_sched_eventmanager(pth_time_t *now, int dopoll)
 
     /* for all threads in the waiting queue... */
     any_occurred = FALSE;
-    for (t = pth_pqueue_head(&pth_WQ); t != NULL; 
+    for (t = pth_pqueue_head(&pth_WQ); t != NULL;
          t = pth_pqueue_walk(&pth_WQ, t, PTH_WALK_NEXT)) {
-    
+
         /* determine signals we block */
         for (sig = 1; sig < PTH_NSIG; sig++)
             if (!sigismember(&(t->mctx.sigs), sig))
@@ -446,7 +459,7 @@ intern void pth_sched_eventmanager(pth_time_t *now, int dopoll)
                 else if (ev->ev_type == PTH_EVENT_SELECT) {
                     /* filedescriptors are checked later all at once.
                        Here we only merge the fd sets. */
-                    pth_util_fds_merge(ev->ev_args.SELECT.nfd, 
+                    pth_util_fds_merge(ev->ev_args.SELECT.nfd,
                                        ev->ev_args.SELECT.rfds, &rfds,
                                        ev->ev_args.SELECT.wfds, &wfds,
                                        ev->ev_args.SELECT.efds, &efds);
@@ -482,7 +495,7 @@ intern void pth_sched_eventmanager(pth_time_t *now, int dopoll)
                 /* Timer */
                 else if (ev->ev_type == PTH_EVENT_TIME) {
                     if (pth_time_cmp(&(ev->ev_args.TIME.tv), now) < 0)
-                        this_occurred = TRUE; 
+                        this_occurred = TRUE;
                     else {
                         /* remember the timer which will be elapsed next */
                         if ((nexttimer_thread == NULL && nexttimer_ev == NULL) ||
@@ -496,52 +509,52 @@ intern void pth_sched_eventmanager(pth_time_t *now, int dopoll)
                 /* Message Port Arrivals */
                 else if (ev->ev_type == PTH_EVENT_MSG) {
                     if (!pth_ring_empty(&(ev->ev_args.MSG.mp->mp_queue)))
-                        this_occurred = TRUE; 
+                        this_occurred = TRUE;
                 }
                 /* Mutex Release */
                 else if (ev->ev_type == PTH_EVENT_MUTEX) {
                     if (!(ev->ev_args.MUTEX.mutex->mx_state & PTH_MUTEX_LOCKED))
-                        this_occurred = TRUE; 
+                        this_occurred = TRUE;
                 }
                 /* Condition Variable Signal */
                 else if (ev->ev_type == PTH_EVENT_COND) {
                     if (ev->ev_args.COND.cond->cn_state & PTH_COND_SIGNALED) {
                         if (ev->ev_args.COND.cond->cn_state & PTH_COND_BROADCAST)
-                            this_occurred = TRUE; 
+                            this_occurred = TRUE;
                         else {
                             if (!(ev->ev_args.COND.cond->cn_state & PTH_COND_HANDLED)) {
                                 ev->ev_args.COND.cond->cn_state |= PTH_COND_HANDLED;
-                                this_occurred = TRUE; 
+                                this_occurred = TRUE;
                             }
                         }
                     }
                 }
                 /* Thread Termination */
                 else if (ev->ev_type == PTH_EVENT_TID) {
-                    if (   (   ev->ev_args.TID.tid == NULL 
+                    if (   (   ev->ev_args.TID.tid == NULL
                             && pth_pqueue_elements(&pth_DQ) > 0)
-                        || (   ev->ev_args.TID.tid != NULL 
+                        || (   ev->ev_args.TID.tid != NULL
                             && ev->ev_args.TID.tid->state == ev->ev_goal))
-                        this_occurred = TRUE; 
-                }
-                /* Process Termination */
-                else if (ev->ev_type == PTH_EVENT_PID) {
-                    pid_t pid;
-                    while (   (pid = pth_sc(waitpid)(ev->ev_args.PID.pid, 
-                                                     ev->ev_args.PID.status, 
-                                                     ev->ev_args.PID.flags|WNOHANG)) < 0
-                           && errno == EINTR)
-                        ;
-                    if (pid > 0)
-                        this_occurred = TRUE; 
+                        this_occurred = TRUE;
                 }
                 /* Custom Event Function */
                 else if (ev->ev_type == PTH_EVENT_FUNC) {
-                    if (ev->ev_args.FUNC.func(ev->ev_args.FUNC.func_arg))
-                        this_occurred = TRUE; 
+                    if (ev->ev_args.FUNC.func(ev->ev_args.FUNC.arg))
+                        this_occurred = TRUE;
+                    else {
+                        pth_time_t tv;
+                        pth_time_set(&tv, now);
+                        pth_time_add(&tv, &(ev->ev_args.FUNC.tv));
+                        if ((nexttimer_thread == NULL && nexttimer_ev == NULL) ||
+                            pth_time_cmp(&tv, &nexttimer_value) < 0) {
+                            nexttimer_thread = t;
+                            nexttimer_ev = ev;
+                            pth_time_set(&nexttimer_value, &tv);
+                        }
+                    }
                 }
 
-                /* tag event when it has occurred */
+                /* tag event if it has occurred */
                 if (this_occurred) {
                     pth_debug2("pth_sched_eventmanager: [non-I/O] event occurred for thread \"%s\"", t->name);
                     ev->ev_occurred = TRUE;
@@ -555,7 +568,7 @@ intern void pth_sched_eventmanager(pth_time_t *now, int dopoll)
 
     /* now decide how to poll for fd I/O and timers */
     if (dopoll) {
-        /* do a polling with immediate timeout, 
+        /* do a polling with immediate timeout,
            i.e. check the fd sets only without blocking */
         pth_time_set(&delay, PTH_TIME_ZERO);
         pdelay = &delay;
@@ -594,12 +607,12 @@ intern void pth_sched_eventmanager(pth_time_t *now, int dopoll)
        handler for signals not catched by events */
     pth_sc(sigprocmask)(SIG_SETMASK, &pth_sigblock, &oss);
 
-    /* now do the polling for filedescriptor I/O and timers 
+    /* now do the polling for filedescriptor I/O and timers
        WHEN THE SCHEDULER SLEEPS AT ALL, THEN HERE!! */
     rc = -1;
     if (!(dopoll && fdmax == -1))
-        while ((rc = pth_sc(select)(fdmax+1, &rfds, &wfds, &efds, pdelay)) < 0 
-               && errno == EINTR) ; 
+        while ((rc = pth_sc(select)(fdmax+1, &rfds, &wfds, &efds, pdelay)) < 0
+               && errno == EINTR) ;
 
     /* restore signal mask and actions and handle signals */
     pth_sc(sigprocmask)(SIG_SETMASK, &oss, NULL);
@@ -607,20 +620,28 @@ intern void pth_sched_eventmanager(pth_time_t *now, int dopoll)
         if (sigismember(&pth_sigcatch, sig))
             sigaction(sig, &osa[sig], NULL);
 
-    /* when the timer elapsed then handle it */
+    /* if the timer elapsed, handle it */
     if (!dopoll && rc == 0 && nexttimer_ev != NULL) {
-        pth_debug2("pth_sched_eventmanager: [timeout] event occurred for thread \"%s\"", 
-                   nexttimer_thread->name);
-        nexttimer_ev->ev_occurred = TRUE;
+        if (nexttimer_ev->ev_type == PTH_EVENT_FUNC) {
+            /* it was an implicit timer event for a function event,
+               so repeat the event handling for rechecking the function */
+            loop_repeat = TRUE;
+        }
+        else {
+            /* it was an explicit timer event, standing for its own */
+            pth_debug2("pth_sched_eventmanager: [timeout] event occurred for thread \"%s\"",
+                       nexttimer_thread->name);
+            nexttimer_ev->ev_occurred = TRUE;
+        }
     }
 
-    /* when the internal signal pipe was used, adjust the select() results */
+    /* if the internal signal pipe was used, adjust the select() results */
     if (!dopoll && rc > 0 && FD_ISSET(pth_sigpipe[0], &rfds)) {
         FD_CLR(pth_sigpipe[0], &rfds);
         rc--;
     }
-    
-    /* when an error occurred, avoid confusion in the cleanup loop */
+
+    /* if an error occurred, avoid confusion in the cleanup loop */
     if (rc <= 0) {
         FD_ZERO(&rfds);
         FD_ZERO(&wfds);
@@ -629,11 +650,11 @@ intern void pth_sched_eventmanager(pth_time_t *now, int dopoll)
 
     /* now comes the final cleanup loop where we've to
        do two jobs: first we've to the late handling of the fd I/O events and
-       additionally when a thread has one occurred event, we move it from the
+       additionally if a thread has one occurred event, we move it from the
        waiting queue to the ready queue */
 
     /* for all threads in the waiting queue... */
-    t = pth_pqueue_head(&pth_WQ); 
+    t = pth_pqueue_head(&pth_WQ);
     while (t != NULL) {
 
         /* do the late handling of the fd I/O and signal
@@ -642,9 +663,9 @@ intern void pth_sched_eventmanager(pth_time_t *now, int dopoll)
         if (t->events != NULL) {
             ev = evh = t->events;
             do {
-                /* 
-                 * Late handling for still not occured events 
-                 */ 
+                /*
+                 * Late handling for still not occured events
+                 */
                 if (!ev->ev_occurred) {
                     /* Filedescriptor I/O */
                     if (ev->ev_type == PTH_EVENT_FD) {
@@ -661,11 +682,11 @@ intern void pth_sched_eventmanager(pth_time_t *now, int dopoll)
                     }
                     /* Filedescriptor Set I/O */
                     else if (ev->ev_type == PTH_EVENT_SELECT) {
-                        if (pth_util_fds_test(ev->ev_args.SELECT.nfd, 
+                        if (pth_util_fds_test(ev->ev_args.SELECT.nfd,
                                               ev->ev_args.SELECT.rfds, &rfds,
                                               ev->ev_args.SELECT.wfds, &wfds,
                                               ev->ev_args.SELECT.efds, &efds)) {
-                            n = pth_util_fds_select(ev->ev_args.SELECT.nfd, 
+                            n = pth_util_fds_select(ev->ev_args.SELECT.nfd,
                                                     ev->ev_args.SELECT.rfds, &rfds,
                                                     ev->ev_args.SELECT.wfds, &wfds,
                                                     ev->ev_args.SELECT.efds, &efds);
@@ -692,7 +713,7 @@ intern void pth_sched_eventmanager(pth_time_t *now, int dopoll)
                         }
                     }
                 }
-                /* 
+                /*
                  * post-processing for already occured events
                  */
                 else {
@@ -722,9 +743,9 @@ intern void pth_sched_eventmanager(pth_time_t *now, int dopoll)
         /* walk to next thread in waiting queue */
         tlast = t;
         t = pth_pqueue_walk(&pth_WQ, t, PTH_WALK_NEXT);
-        
-        /* 
-         * move last thread to ready queue when any events occurred for it.
+
+        /*
+         * move last thread to ready queue if any events occurred for it.
          * we insert it with a slightly increased queue priority to it a
          * better chance to immediately get scheduled, else the last running
          * thread might immediately get again the CPU which is usually not
@@ -740,6 +761,10 @@ intern void pth_sched_eventmanager(pth_time_t *now, int dopoll)
         }
     }
 
+    /* perhaps we have to internally loop... */
+    if (loop_repeat)
+        goto loop_entry;
+
     pth_debug1("pth_sched_eventmanager: leaving");
     return;
 }
