@@ -1,6 +1,6 @@
 /*
 **  GNU Pth - The GNU Portable Threads
-**  Copyright (c) 1999 Ralf S. Engelschall <rse@engelschall.com>
+**  Copyright (c) 1999-2000 Ralf S. Engelschall <rse@engelschall.com>
 **
 **  This file is part of GNU Pth, a non-preemptive thread scheduling
 **  library which can be found at http://www.gnu.org/software/pth/.
@@ -22,7 +22,9 @@
 **
 **  test_misc.c: Pth test program (misc functions)
 */
-
+                             /* ``Study it forever and you'll still wonder.
+                                  Fly it once and you'll know.''
+                                                 -- Henry Spencer  */
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -82,7 +84,7 @@ static void *my_child(void *_arg)
     return NULL;
 }
 
-int main(int argc, char *argv[]) 
+int main(int argc, char *argv[])
 {
     pth_t child[10];
     pth_attr_t t_attr;
